==> ?url=gopher%3A%2F%2Fftp.icm.edu.pl%3A70%2F9%2Fvol%2Frzm8%2Fzebra%2Ficmarch%2Fzebra-0.71.tar.gz: zebra-0.71/bgpd/bgp_clist.c <==
/*
 * BGP Community list.
 * Copyright (C) 1999 Kunihiro Ishiguro
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>

#include "command.h"
#include "memory.h"
#include "log.h"
#include "buffer.h"

#include "bgpd/bgpd.h"
#include "bgpd/bgp_community.h"
#include "bgpd/bgp_clist.h"

enum community_entry_type
{
  COMMUNITY_DENY,
  COMMUNITY_PERMIT
};

struct community_list_list
{
  struct community_list *head;
  struct community_list *tail;
};

struct community_list_master
{
  struct community_list_list num;
  struct community_list_list str;
};

struct community_entry
{
  struct community_entry *next;
  struct community_entry *prev;

  enum community_entry_type type;

  struct community *com;
};

static struct community_list_master community_list_master =
{
  {NULL, NULL},
  {NULL, NULL}
};

struct community_entry *
community_entry_new ()
{
  struct community_entry *new;

  new = XMALLOC (MTYPE_COMMUNITY_ENTRY, sizeof (struct community_entry));
  bzero (new, sizeof (struct community_entry));
  return new;
}

void
community_entry_free (struct community_entry *entry)
{
  if (entry->com)
    community_free (entry->com);
  XFREE (MTYPE_COMMUNITY_ENTRY, entry);
}

struct community_entry *
community_entry_make (struct community *com, enum community_entry_type type)
{
  struct community_entry *entry;

  entry = community_entry_new ();
  entry->com = com;
  entry->type = type;

  return entry;
}

struct community_list *
community_list_new ()
{
  struct community_list *new;

  new = XMALLOC (MTYPE_COMMUNITY_LIST, sizeof (struct community_list));
  bzero (new, sizeof (struct community_list));
  return new;
}

void
community_list_free (struct community_list *list)
{
  if (list->name)
    XFREE (0, list->name);
  XFREE (MTYPE_COMMUNITY_LIST, list);
}

struct community_list *
community_list_insert (char *name)
{
  int i;
  long number;
  struct community_list *new;
  struct community_list *point;
  struct community_list_list *list;

  /* Allocate new community_list and copy given name. */
  new = community_list_new ();
  new->name = strdup (name);

  /* If name is made by all digit character.  We treat it as
     number. */
  for (number = 0, i = 0; i < strlen (name); i++)
    {
      if (isdigit (name[i]))
	number = (number * 10) + (name[i] - '0');
      else
	break;
    }

  /* In case of name is all digit character */
  if (i == strlen (name))
    {
      new->sort = COMMUNITY_LIST_NUMBER;

      /* Set access_list to number list. */
      list = &community_list_master.num;

      for (point = list->head; point; point = point->next)
	if (atol (point->name) >= number)
	  break;
    }
  else
    {
      new->sort = COMMUNITY_LIST_STRING;

      /* Set access_list to string list. */
      list = &community_list_master.str;
  
      /* Set point to insertion point. */
      for (point = list->head; point; point = point->next)
	if (strcmp (point->name, name) >= 0)
	  break;
    }

  /* In case of this is the first element of master. */
  if (list->head == NULL)
    {
      list->head = list->tail = new;
      return new;
    }

  /* In case of insertion is made at the tail of access_list. */
  if (point == NULL)
    {
      new->prev = list->tail;
      list->tail->next = new;
      list->tail = new;
      return new;
    }

  /* In case of insertion is made at the head of access_list. */
  if (point == list->head)
    {
      new->next = list->head;
      list->head->prev = new;
      list->head = new;
      return new;
    }

  /* Insertion is made at middle of the access_list. */
  new->next = point;
  new->prev = point->prev;

  if (point->prev)
    point->prev->next = new;
  point->prev = new;

  return new;
}

struct community_list *
community_list_lookup (char *name)
{
  struct community_list *list;

  if (name == NULL)
    return NULL;

  for (list = community_list_master.num.head; list; list = list->next)
    if (strcmp (list->name, name) == 0)
      return list;

  for (list = community_list_master.str.head; list; list = list->next)
    if (strcmp (list->name, name) == 0)
      return list;

  return NULL;
}

struct community_list *
community_list_get (char *name)
{
  struct community_list *list;

  list = community_list_lookup (name);
  if (list == NULL)
    list = community_list_insert (name);
  return list;
}

struct community_entry *
community_entry_lookup (struct community_list *list, 
			struct community *com, enum community_entry_type type)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    if (community_cmp (entry->com, com))
      return entry;
  return NULL;
}

void
community_list_entry_add (struct community_list *list, 
			  struct community_entry *entry)
{
  entry->next = NULL;
  entry->prev = list->tail;

  if (list->tail)
    list->tail->next = entry;
  else
    list->head = entry;
  list->tail = entry;
}

void
community_list_delete (struct community_list *list)
{
  struct community_list_list *clist;

  if (list->sort == COMMUNITY_LIST_NUMBER)
    clist = &community_list_master.num;
  else
    clist = &community_list_master.str;

  if (list->next)
    list->next->prev = list->prev;
  else
    clist->tail = list->prev;

  if (list->prev)
    list->prev->next = list->next;
  else
    clist->head = list->next;

  community_list_free (list);
}

int 
community_list_empty (struct community_list *list)
{
  if (list->head == NULL && list->tail == NULL)
    return 1;
  else
    return 0;
}

void
community_list_entry_delete (struct community_list *list,
			     struct community_entry *entry)
{
  if (entry->next)
    entry->next->prev = entry->prev;
  else
    list->tail = entry->prev;

  if (entry->prev)
    entry->prev->next = entry->next;
  else
    list->head = entry->next;

  community_entry_free (entry);

  if (community_list_empty (list))
    community_list_delete (list);
}

int
community_list_match (struct community *com, struct community_list *list)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    if (community_match (com, entry->com))
      return 1;
  return 0;
}

char *
community_type_str (enum community_entry_type type)
{
  switch (type)
    {
    case COMMUNITY_DENY:
      return "deny";
      break;
    case COMMUNITY_PERMIT:
      return "permit";
      break;
    default:
      return "";
      break;
    }
}

void
community_list_print (struct community_list *list)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    printf ("ip community-list %s %s%s\n", list->name, 
	    community_type_str (entry->type), community_print (entry->com));
}

DEFUN (ip_community_list, ip_community_list_cmd,
       "ip community-list NAME (deny|permit) ...",
       IP_STR
       "Community attribute list.\n"
       "Community list name\n"
       "Community list for denies\n"
       "Community list for permits\n"
       "Community list\n")
{
  enum community_entry_type type;
  struct community_entry *entry;
  struct community_list *list;
  struct community *com;
  struct buffer *b;
  int i;
  char *str;
  int first = 0;

  /* Check the list type. */
  if (strcmp (argv[1], "permit") == 0)
    type = COMMUNITY_PERMIT;
  else if (strcmp (argv[1], "deny") == 0)
    type = COMMUNITY_DENY;
  else
    {
      vty_out (vty, "community-list type must be [permit|deny]\r\n");
      return CMD_WARNING;
    }

  b = buffer_new (BUFFER_STRING, 1024);
  for (i = 2; i < argc; i++)
    {
      if (first)
	buffer_putc (b, ' ');
      else
	first = 1;

      buffer_putstr (b, argv[i]);
    }
  buffer_putc (b, '\0');

  str = buffer_getstr (b);
  buffer_free (b);

  com = community_str2com (str);
  if (! com)
    {
      vty_out (vty, "Community-list malformed: %s\r\n", str);
      free (str);
      return CMD_WARNING;
    }

  entry = community_entry_make (com, type);
  
  free (str);

  /* Install new community list to the community_list. */
  list = community_list_get (argv[0]);
  community_list_entry_add (list, entry);

  return CMD_SUCCESS;
}

DEFUN (no_ip_community_list, no_ip_community_list_cmd,
       "no ip community-list NAME (deny|permit) ...",
       NO_STR
       IP_STR
       "Community attribute list.\n"
       "Community list name\n"
       "Community list for denies\n"
       "Community list for permits\n"
       "Community list\n")
{
  enum community_entry_type type;
  struct community_entry *entry;
  struct community_list *list;
  struct community *com;
  struct buffer *b;
  int i;
  char *str;
  int first = 0;

  list = community_list_lookup (argv[0]);
  if (list == NULL)
    {
      vty_out (vty, "ip community-list %s doesn't exist.\r\n", argv[0]);
      return CMD_WARNING;
    }

  /* Check the list type. */
  if (strcmp (argv[1], "permit") == 0)
    type = COMMUNITY_PERMIT;
  else if (strcmp (argv[1], "deny") == 0)
    type = COMMUNITY_DENY;
  else
    {
      vty_out (vty, "community-list type must be [permit|deny]\r\n");
      return CMD_WARNING;
    }

  b = buffer_new (BUFFER_STRING, 1024);
  for (i = 2; i < argc; i++)
    {
      if (first)
	buffer_putc (b, ' ');
      else
	first = 1;

      buffer_putstr (b, argv[i]);
    }
  buffer_putc (b, '\0');

  str = buffer_getstr (b);
  buffer_free (b);

  com = community_str2com (str);
  free (str);

  if (! com)
    {
      vty_out (vty, "Community-list malformed: %s\r\n", str);
      return CMD_WARNING;
    }

  entry = community_entry_lookup (list, com, type);

  if (entry == NULL)
    {
      vty_out (vty, "Can't find specified community list.\r\n");
      return CMD_WARNING;
    }

  community_list_entry_delete (list, entry);

  return CMD_SUCCESS;
}

int
config_write_community (struct vty *vty)
{
  struct community_list *list;
  struct community_entry *entry;
  int write = 0;

  for (list = community_list_master.num.head; list; list = list->next)
    for (entry = list->head; entry; entry = entry->next)
      {
	vty_out (vty, "ip community-list %s %s%s%s",
		 list->name, community_type_str (entry->type), 
		 community_print (entry->com), VTY_NEWLINE);
	write++;
      }

  for (list = community_list_master.str.head; list; list = list->next)
    for (entry = list->head; entry; entry = entry->next)
      {
	vty_out (vty, "ip community-list %s %s%s%s",
		 list->name, community_type_str (entry->type), 
		 community_print (entry->com), VTY_NEWLINE);
	write++;
      }
  return write;
}

struct cmd_node community_list_node =
{
  COMMUNITY_LIST_NODE,
  ""
};

void
community_list_init ()
{
  install_node (&community_list_node, config_write_community);

  install_element (CONFIG_NODE, &ip_community_list_cmd);
  install_element (CONFIG_NODE, &no_ip_community_list_cmd);
}

void
community_list_test ()
{
  struct community *com1;
  struct community *com2;
  struct community_list *list;
  struct community_entry *entry;

  /* ip community-list 10 permit no-export. */
  list = community_list_get ("10");

  com1 = community_str2com ("7675:70");
  entry = community_entry_make (com1, COMMUNITY_PERMIT);
  community_list_entry_add (list, entry);

  com1 = community_str2com ("7675:60");
  entry = community_entry_make (com1, COMMUNITY_PERMIT);
  community_list_entry_add (list, entry);

  com2 = community_str2com ("7675:60");
  entry = community_entry_lookup (list, com2, COMMUNITY_PERMIT);
  community_list_entry_delete (list, entry);

  community_list_print (list);
}

==> ?url=gopher%3A%2F%2Fftp.icm.edu.pl%3A70%2F9%2Fvol%2Frzm8%2Fzebra%2Ficmarch%2Fzebra-0.75.tar.gz: zebra-0.75/bgpd/bgp_clist.c <==
/*
 * BGP Community list.
 * Copyright (C) 1999 Kunihiro Ishiguro
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>

#include "command.h"
#include "memory.h"
#include "log.h"
#include "buffer.h"

#include "bgpd/bgpd.h"
#include "bgpd/bgp_community.h"
#include "bgpd/bgp_clist.h"

enum community_entry_type
{
  COMMUNITY_DENY,
  COMMUNITY_PERMIT
};

struct community_list_list
{
  struct community_list *head;
  struct community_list *tail;
};

struct community_list_master
{
  struct community_list_list num;
  struct community_list_list str;
};

struct community_entry
{
  struct community_entry *next;
  struct community_entry *prev;

  enum community_entry_type type;

  struct community *com;
};

static struct community_list_master community_list_master =
{
  {NULL, NULL},
  {NULL, NULL}
};

struct community_entry *
community_entry_new ()
{
  struct community_entry *new;

  new = XMALLOC (MTYPE_COMMUNITY_ENTRY, sizeof (struct community_entry));
  bzero (new, sizeof (struct community_entry));
  return new;
}

void
community_entry_free (struct community_entry *entry)
{
  if (entry->com)
    community_free (entry->com);
  XFREE (MTYPE_COMMUNITY_ENTRY, entry);
}

struct community_entry *
community_entry_make (struct community *com, enum community_entry_type type)
{
  struct community_entry *entry;

  entry = community_entry_new ();
  entry->com = com;
  entry->type = type;

  return entry;
}

struct community_list *
community_list_new ()
{
  struct community_list *new;

  new = XMALLOC (MTYPE_COMMUNITY_LIST, sizeof (struct community_list));
  bzero (new, sizeof (struct community_list));
  return new;
}

void
community_list_free (struct community_list *list)
{
  if (list->name)
    XFREE (0, list->name);
  XFREE (MTYPE_COMMUNITY_LIST, list);
}

struct community_list *
community_list_insert (char *name)
{
  int i;
  long number;
  struct community_list *new;
  struct community_list *point;
  struct community_list_list *list;

  /* Allocate new community_list and copy given name. */
  new = community_list_new ();
  new->name = strdup (name);

  /* If name is made by all digit character.  We treat it as
     number. */
  for (number = 0, i = 0; i < strlen (name); i++)
    {
      if (isdigit (name[i]))
	number = (number * 10) + (name[i] - '0');
      else
	break;
    }

  /* In case of name is all digit character */
  if (i == strlen (name))
    {
      new->sort = COMMUNITY_LIST_NUMBER;

      /* Set access_list to number list. */
      list = &community_list_master.num;

      for (point = list->head; point; point = point->next)
	if (atol (point->name) >= number)
	  break;
    }
  else
    {
      new->sort = COMMUNITY_LIST_STRING;

      /* Set access_list to string list. */
      list = &community_list_master.str;
  
      /* Set point to insertion point. */
      for (point = list->head; point; point = point->next)
	if (strcmp (point->name, name) >= 0)
	  break;
    }

  /* In case of this is the first element of master. */
  if (list->head == NULL)
    {
      list->head = list->tail = new;
      return new;
    }

  /* In case of insertion is made at the tail of access_list. */
  if (point == NULL)
    {
      new->prev = list->tail;
      list->tail->next = new;
      list->tail = new;
      return new;
    }

  /* In case of insertion is made at the head of access_list. */
  if (point == list->head)
    {
      new->next = list->head;
      list->head->prev = new;
      list->head = new;
      return new;
    }

  /* Insertion is made at middle of the access_list. */
  new->next = point;
  new->prev = point->prev;

  if (point->prev)
    point->prev->next = new;
  point->prev = new;

  return new;
}

struct community_list *
community_list_lookup (char *name)
{
  struct community_list *list;

  if (name == NULL)
    return NULL;

  for (list = community_list_master.num.head; list; list = list->next)
    if (strcmp (list->name, name) == 0)
      return list;

  for (list = community_list_master.str.head; list; list = list->next)
    if (strcmp (list->name, name) == 0)
      return list;

  return NULL;
}

struct community_list *
community_list_get (char *name)
{
  struct community_list *list;

  list = community_list_lookup (name);
  if (list == NULL)
    list = community_list_insert (name);
  return list;
}

struct community_entry *
community_entry_lookup (struct community_list *list, 
			struct community *com, enum community_entry_type type)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    if (community_cmp (entry->com, com))
      return entry;
  return NULL;
}

void
community_list_entry_add (struct community_list *list, 
			  struct community_entry *entry)
{
  entry->next = NULL;
  entry->prev = list->tail;

  if (list->tail)
    list->tail->next = entry;
  else
    list->head = entry;
  list->tail = entry;
}

void
community_list_delete (struct community_list *list)
{
  struct community_list_list *clist;

  if (list->sort == COMMUNITY_LIST_NUMBER)
    clist = &community_list_master.num;
  else
    clist = &community_list_master.str;

  if (list->next)
    list->next->prev = list->prev;
  else
    clist->tail = list->prev;

  if (list->prev)
    list->prev->next = list->next;
  else
    clist->head = list->next;

  community_list_free (list);
}

int 
community_list_empty (struct community_list *list)
{
  if (list->head == NULL && list->tail == NULL)
    return 1;
  else
    return 0;
}

void
community_list_entry_delete (struct community_list *list,
			     struct community_entry *entry)
{
  if (entry->next)
    entry->next->prev = entry->prev;
  else
    list->tail = entry->prev;

  if (entry->prev)
    entry->prev->next = entry->next;
  else
    list->head = entry->next;

  community_entry_free (entry);

  if (community_list_empty (list))
    community_list_delete (list);
}

int
community_list_match (struct community *com, struct community_list *list)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    if (community_match (com, entry->com))
      return 1;
  return 0;
}

char *
community_type_str (enum community_entry_type type)
{
  switch (type)
    {
    case COMMUNITY_DENY:
      return "deny";
      break;
    case COMMUNITY_PERMIT:
      return "permit";
      break;
    default:
      return "";
      break;
    }
}

void
community_list_print (struct community_list *list)
{
  struct community_entry *entry;

  for (entry = list->head; entry; entry = entry->next)
    printf ("ip community-list %s %s%s\n", list->name, 
	    community_type_str (entry->type), community_print (entry->com));
}

DEFUN (ip_community_list, ip_community_list_cmd,
       "ip community-list NAME (deny|permit) .COMMUNITY",
       IP_STR
       "Community attribute list\n"
       "Community list name\n"
       "Community list for denies\n"
       "Community list for permits\n"
       "Community list. e.g. no-export 7675:70\n")
{
  enum community_entry_type type;
  struct community_entry *entry;
  struct community_list *list;
  struct community *com;
  struct buffer *b;
  int i;
  char *str;
  int first = 0;

  /* Check the list type. */
  if (strcmp (argv[1], "permit") == 0)
    type = COMMUNITY_PERMIT;
  else if (strcmp (argv[1], "deny") == 0)
    type = COMMUNITY_DENY;
  else
    {
      vty_out (vty, "community-list type must be [permit|deny]\r\n");
      return CMD_WARNING;
    }

  b = buffer_new (BUFFER_STRING, 1024);
  for (i = 2; i < argc; i++)
    {
      if (first)
	buffer_putc (b, ' ');
      else
	first = 1;

      buffer_putstr (b, argv[i]);
    }
  buffer_putc (b, '\0');

  str = buffer_getstr (b);
  buffer_free (b);

  com = community_str2com (str);
  if (! com)
    {
      vty_out (vty, "Community-list malformed: %s\r\n", str);
      free (str);
      return CMD_WARNING;
    }

  entry = community_entry_make (com, type);
  
  free (str);

  /* Install new community list to the community_list. */
  list = community_list_get (argv[0]);
  community_list_entry_add (list, entry);

  return CMD_SUCCESS;
}

DEFUN (no_ip_community_list, no_ip_community_list_cmd,
       "no ip community-list NAME (deny|permit) .COMMUNITY",
       NO_STR
       IP_STR
       "Community attribute list\n"
       "Community list name\n"
       "Community list for denies\n"
       "Community list for permits\n"
       "Community list. e.g no-export 7675:70\n")
{
  enum community_entry_type type;
  struct community_entry *entry;
  struct community_list *list;
  struct community *com;
  struct buffer *b;
  int i;
  char *str;
  int first = 0;

  list = community_list_lookup (argv[0]);
  if (list == NULL)
    {
      vty_out (vty, "ip community-list %s doesn't exist.\r\n", argv[0]);
      return CMD_WARNING;
    }

  /* Check the list type. */
  if (strcmp (argv[1], "permit") == 0)
    type = COMMUNITY_PERMIT;
  else if (strcmp (argv[1], "deny") == 0)
    type = COMMUNITY_DENY;
  else
    {
      vty_out (vty, "community-list type must be [permit|deny]\r\n");
      return CMD_WARNING;
    }

  b = buffer_new (BUFFER_STRING, 1024);
  for (i = 2; i < argc; i++)
    {
      if (first)
	buffer_putc (b, ' ');
      else
	first = 1;

      buffer_putstr (b, argv[i]);
    }
  buffer_putc (b, '\0');

  str = buffer_getstr (b);
  buffer_free (b);

  com = community_str2com (str);
  free (str);

  if (! com)
    {
      vty_out (vty, "Community-list malformed: %s\r\n", str);
      return CMD_WARNING;
    }

  entry = community_entry_lookup (list, com, type);

  if (entry == NULL)
    {
      vty_out (vty, "Can't find specified community list.\r\n");
      return CMD_WARNING;
    }

  community_list_entry_delete (list, entry);

  return CMD_SUCCESS;
}

int
config_write_community (struct vty *vty)
{
  struct community_list *list;
  struct community_entry *entry;
  int write = 0;

  for (list = community_list_master.num.head; list; list = list->next)
    for (entry = list->head; entry; entry = entry->next)
      {
	vty_out (vty, "ip community-list %s %s%s%s",
		 list->name, community_type_str (entry->type), 
		 community_print (entry->com), VTY_NEWLINE);
	write++;
      }

  for (list = community_list_master.str.head; list; list = list->next)
    for (entry = list->head; entry; entry = entry->next)
      {
	vty_out (vty, "ip community-list %s %s%s%s",
		 list->name, community_type_str (entry->type), 
		 community_print (entry->com), VTY_NEWLINE);
	write++;
      }
  return write;
}

struct cmd_node community_list_node =
{
  COMMUNITY_LIST_NODE,
  ""
};

void
community_list_init ()
{
  install_node (&community_list_node, config_write_community);

  install_element (CONFIG_NODE, &ip_community_list_cmd);
  install_element (CONFIG_NODE, &no_ip_community_list_cmd);
}

void
community_list_test ()
{
  struct community *com1;
  struct community *com2;
  struct community_list *list;
  struct community_entry *entry;

  /* ip community-list 10 permit no-export. */
  list = community_list_get ("10");

  com1 = community_str2com ("7675:70");
  entry = community_entry_make (com1, COMMUNITY_PERMIT);
  community_list_entry_add (list, entry);

  com1 = community_str2com ("7675:60");
  entry = community_entry_make (com1, COMMUNITY_PERMIT);
  community_list_entry_add (list, entry);

  com2 = community_str2com ("7675:60");
  entry = community_entry_lookup (list, com2, COMMUNITY_PERMIT);
  community_list_entry_delete (list, entry);

  community_list_print (list);
}
